# test_history_undo.py
# Tests of the history undo command.
#
# Copyright (C) 2013-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

"""Tests of the history undo command."""

from __future__ import absolute_import
from __future__ import unicode_literals
from dnf.exceptions import PackagesNotAvailableError, PackagesNotInstalledError
from dnf.history import NEVRAOperations
from dnf.package import Package
from dnf.transaction import (ERASE, DOWNGRADE, INSTALL, REINSTALL,
                             TransactionItem, UPGRADE)
from hawkey import split_nevra
from tests.support import mock_sack, Base, ObjectMatcher
from unittest import TestCase

class BaseTest(TestCase):
    """Unit tests of dnf.Base."""

    def _create_item_matcher(self, op_type, installed=None, erased=None,
                             obsoleted=[], reason='unknown'):
        """Create a new instance of dnf.transaction.TransactionItem matcher."""
        attrs = {'op_type': op_type,
                 'installed': self._create_package_matcher(installed)
                              if installed else installed,
                 'erased': self._create_package_matcher(erased)
                           if erased else erased,
                 'obsoleted': [self._create_package_matcher(nevra)
                               for nevra in obsoleted],
                 'reason': reason}
        return ObjectMatcher(TransactionItem, attrs)

    def _create_package_matcher(self, nevra_str):
        """Create a new instance of dnf.package.Package matcher."""
        nevra = split_nevra(nevra_str)
        attrs = {'name': nevra.name,
                 'epoch': nevra.epoch,
                 'version': nevra.version,
                 'release': nevra.release,
                 'arch': nevra.arch}
        return ObjectMatcher(Package, attrs)

    def setUp(self):
        """Prepare the test fixture."""
        self._base = Base()
        self._base._sack = mock_sack('main', 'updates')

    def test_history_undo_operations_downgrade(self):
        """Test history_undo_operations with a downgrade."""
        operations = NEVRAOperations()
        operations.add('Downgrade', 'pepper-20-0.x86_64', 'pepper-20-1.x86_64', ('lotus-3-16.x86_64',))

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                UPGRADE, installed='pepper-20-1.x86_64', erased='pepper-20-0.x86_64'))
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                INSTALL, installed='lotus-3-16.x86_64', reason='user'))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_downgrade_notavailable(self):
        """Test history_undo_operations with an unavailable downgrade."""
        operations = NEVRAOperations()
        operations.add('Downgrade', 'pepper-20-0.x86_64', 'pepper-20-2.x86_64')

        with self._base, self.assertRaises(PackagesNotAvailableError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'pepper-20-2.x86_64')

    def test_history_undo_operations_downgrade_notinstalled(self):
        """Test history_undo_operations with a not installed downgrade."""
        operations = NEVRAOperations()
        operations.add('Downgrade', 'lotus-3-0.x86_64', 'lotus-3-16.x86_64')

        with self._base, self.assertRaises(PackagesNotInstalledError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'lotus-3-0.x86_64')

    def test_history_undo_operations_erase(self):
        """Test history_undo_operations with an erase."""
        operations = NEVRAOperations()
        operations.add('Erase', 'lotus-3-16.x86_64')

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                INSTALL, installed='lotus-3-16.x86_64', reason='user'))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_erase_twoavailable(self):
        """Test history_undo_operations with an erase available in two repos."""
        base = Base()
        base._sack = mock_sack('main', 'search')
        operations = NEVRAOperations()
        operations.add('Erase', 'lotus-3-16.x86_64')

        with base:
            base._history_undo_operations(operations, 0)

            transaction_it = iter(base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                INSTALL, installed='lotus-3-16.x86_64', reason='user'))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_erase_notavailable(self):
        """Test history_undo_operations with an unavailable erase."""
        operations = NEVRAOperations()
        operations.add('Erase', 'hole-1-1.x86_64')

        with self._base, self.assertRaises(PackagesNotAvailableError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'hole-1-1.x86_64')

    def test_history_undo_operations_install(self):
        """Test history_undo_operations with an install."""
        operations = NEVRAOperations()
        operations.add('Install', 'pepper-20-0.x86_64', obsoleted_nevras=('lotus-3-16.x86_64',))

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                ERASE, erased='pepper-20-0.x86_64'))
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                INSTALL, installed='lotus-3-16.x86_64', reason='user'))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_install_notinstalled(self):
        """Test history_undo_operations with a not installed install."""
        operations = NEVRAOperations()
        operations.add('Install', 'mrkite-2-0.x86_64')

        with self._base, self.assertRaises(PackagesNotInstalledError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'mrkite-2-0.x86_64')

    def test_history_undo_operations_reinstall(self):
        """Test history_undo_operations with a reinstall."""
        operations = NEVRAOperations()
        operations.add('Reinstall', 'pepper-20-0.x86_64', 'pepper-20-0.x86_64', ('hole-1-1.x86_64',))

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                REINSTALL, installed='pepper-20-0.x86_64', erased='pepper-20-0.x86_64',
                obsoleted=('hole-1-1.x86_64',)))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_reinstall_notavailable(self):
        """Test history_undo_operations with an unvailable reinstall."""
        operations = NEVRAOperations()
        operations.add('Reinstall', 'mrkite-2-0.x86_64', 'mrkite-2-0.x86_64')

        with self._base, self.assertRaises(PackagesNotInstalledError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'mrkite-2-0.x86_64')

    def test_history_undo_operations_reinstall_notinstalled(self):
        """Test history_undo_operations with a not installed reinstall."""
        operations = NEVRAOperations()
        operations.add('Reinstall', 'hole-1-1.x86_64', 'hole-1-1.x86_64')

        with self._base, self.assertRaises(PackagesNotAvailableError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'hole-1-1.x86_64')

    def test_history_undo_operations_reinstall_notinstalled_obsoleted(self):
        """Test history_undo_operations with a not installed obsoleted of a reinstall."""
        operations = NEVRAOperations()
        operations.add('Reinstall', 'pepper-20-0.x86_64', 'pepper-20-0.x86_64', ('lotus-3-16.x86_64',))

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                REINSTALL, installed='pepper-20-0.x86_64', erased='pepper-20-0.x86_64',
                obsoleted=()))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_update(self):
        """Test history_undo_operations with an update."""
        operations = NEVRAOperations()
        operations.add('Update', 'tour-5-0.noarch', 'tour-4.6-1.noarch', ('lotus-3-16.x86_64',))

        with self._base:
            self._base._history_undo_operations(operations, 0)

            transaction_it = iter(self._base.transaction)
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                DOWNGRADE, installed='tour-4.6-1.noarch', erased='tour-5-0.noarch'))
            self.assertEqual(next(transaction_it), self._create_item_matcher(
                INSTALL, installed='lotus-3-16.x86_64', reason='user'))
            self.assertRaises(StopIteration, next, transaction_it)

    def test_history_undo_operations_update_notavailable(self):
        """Test history_undo_operations with an unavailable update."""
        operations = NEVRAOperations()
        operations.add('Update', 'tour-5-0.noarch', 'tour-4.6-2.noarch')

        with self._base, self.assertRaises(PackagesNotAvailableError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'tour-4.6-2.noarch')

    def test_history_undo_operations_update_notinstalled(self):
        """Test history_undo_operations with a not installed update."""
        operations = NEVRAOperations()
        operations.add('Update', 'lotus-4-0.x86_64', 'lotus-3-16.x86_64')

        with self._base, self.assertRaises(PackagesNotInstalledError) as context:
            self._base._history_undo_operations(operations, 0)

        self.assertEqual(context.exception.pkg_spec, 'lotus-4-0.x86_64')



# Copyright (C) 2014-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
import dnf.goal
import dnf.selector
import tests.support


class GoalTest(tests.support.TestCase):
    def setUp(self):
        base = tests.support.MockBase('main')
        self.sack = base.sack
        self.goal = dnf.goal.Goal(self.sack)

    def test_get_reason(self):
        sltr = dnf.selector.Selector(self.sack)
        sltr.set(name='mrkite')
        grp_sltr = dnf.selector.Selector(self.sack)
        grp_sltr.set(name='lotus')

        goal = self.goal
        goal.install(select=sltr)
        goal.install(select=grp_sltr)
        goal.group_members.add('lotus')
        goal.run()
        installs = goal.list_installs()
        mrkite = [pkg for pkg in installs if pkg.name == 'mrkite'][0]
        lotus = [pkg for pkg in installs if pkg.name == 'lotus'][0]
        trampoline = [pkg for pkg in installs if pkg.name == 'trampoline'][0]
        self.assertEqual(goal.get_reason(lotus), 'group')
        self.assertEqual(goal.get_reason(mrkite), 'user')
        self.assertEqual(goal.get_reason(trampoline), 'dep')

    def test_group_reason(self):
        goal = self.goal
        hole = self.sack.query().filter(name='hole')[0]
        goal.group_members.add('hole')
        self.assertEqual('group', goal.group_reason(hole, 'unknown'))
        self.assertEqual('dep', goal.group_reason(hole, 'dep'))

    def test_push_userinstalled(self):
        base = tests.support.MockBase('main')
        base.conf.clean_requirements_on_remove = True
        goal = self.goal
        installed = base.sack.query().installed()
        for pkg in installed:
            base._yumdb.get_package(pkg).reason = 'dep'
        pkg1 = installed.filter(name="pepper")[0]
        base._yumdb.get_package(pkg1).reason = "user"
        pkg2 = installed.filter(name="hole")[0]
        base._yumdb.get_package(pkg2).reason = "unknown"
        pkgs = installed.filter(name__neq=["pepper", "hole", "librita"]
                               ).run()

        # test:
        goal.push_userinstalled(installed, base._yumdb)
        goal.run()
        self.assertEqual(goal.list_unneeded(), pkgs)



#!/usr/bin/python


import os
import json

import modulemd


MODULES_DIR = os.path.join(os.path.dirname(__file__), "..", "modules")
SPECS_DIR = os.path.join(os.path.dirname(__file__), "..", "specs")


def parse_module_id(module_id):
    return module_id.rsplit("-", 2)


for module_id in os.listdir(MODULES_DIR):
    if module_id.startswith("_"):
        continue

    name, stream, version = parse_module_id(module_id)

    profiles_file = os.path.join(SPECS_DIR, module_id, "profiles.json")
    if os.path.isfile(profiles_file):
        with open(profiles_file, "r") as f:
            profiles = json.load(f)
    else:
        profiles = {}

    for arch in os.listdir(os.path.join(MODULES_DIR, module_id)):
        if arch == "noarch":
            continue

        module_dir = os.path.join(MODULES_DIR, module_id, arch)
        rpms = [i for i in os.listdir(module_dir) if i.endswith(".rpm")]

        noarch_module_dir = os.path.join(MODULES_DIR, module_id, "noarch")
        if os.path.isdir(noarch_module_dir):
            noarch_rpms = [i for i in os.listdir(noarch_module_dir) if i.endswith(".rpm")]
        else:
            noarch_rpms = []

        rpms = sorted(set(rpms) | set(noarch_rpms))

        # HACK: force epoch to make test data compatible with libmodulemd >= 1.4.0
        rpms_with_epoch = []
        for i in rpms:
            n, v, ra = i.rsplit("-", 2)
            nevra = "%s-0:%s-%s" % (n, v, ra)
            rpms_with_epoch.append(nevra)
        rpms = rpms_with_epoch

        mmd = modulemd.ModuleMetadata()
        mmd.name = name
        mmd.stream = stream
        mmd.version = int(version)
        mmd.add_module_license("LGPLv2")
        mmd.summary = "Fake module"
        mmd.description = mmd.summary
        for rpm in rpms:
            mmd.components.add_rpm(rpm.rsplit("-", 2)[0], "")
            mmd.artifacts.add_rpm(rpm[:-4])
        for profile_name in profiles:
            profile = modulemd.ModuleProfile()
            profile.rpms.update(profiles[profile_name]["rpms"])
            mmd.profiles[profile_name] = profile

        mmd.dump(os.path.join(module_dir, "%s.%s.yaml" % (module_id, arch)))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import
from __future__ import unicode_literals
from . import misc
import dnf.i18n
import dnf.pycomp
import glob
import logging
import os
import rpm

logger = logging.getLogger('dnf')

# For returnPackages(patterns=)
flags = {"GT": rpm.RPMSENSE_GREATER,
         "GE": rpm.RPMSENSE_EQUAL | rpm.RPMSENSE_GREATER,
         "LT": rpm.RPMSENSE_LESS,
         "LE": rpm.RPMSENSE_LESS | rpm.RPMSENSE_EQUAL,
         "EQ": rpm.RPMSENSE_EQUAL,
         None: 0 }


def _open_no_umask(*args):
    """ Annoying people like to set umask's for root, which screws everything
        up for user readable stuff. """
    oumask = os.umask(0o22)
    try:
        if dnf.pycomp.PY3:
            ret = open(*args, encoding='utf-8')
        else:
            ret = open(*args)
    finally:
        os.umask(oumask)

    return ret


def _makedirs_no_umask(*args):
    """ Annoying people like to set umask's for root, which screws everything
        up for user readable stuff. """
    oumask = os.umask(0o22)
    try:
        ret = os.makedirs(*args)
    finally:
        os.umask(oumask)

    return ret


def _iopen(*args):
    """ IOError wrapper BS for open, stupid exceptions. """
    try:
        if dnf.pycomp.PY3:
            ret = open(*args, encoding='utf-8')
        else:
            ret = open(*args)
    except IOError as e:
        return None, e
    return ret, None


def _sanitize(path):
    return path.replace('/', '').replace('~', '')


class AdditionalPkgDB(object):
    """ Accesses additional package data rpmdb is unable to store.

        Previously known as yumdb.
    """
    # dir: <persistdir>/yumdb
    # pkgs stored in name[0]/pkgid-name-ver-rel-arch dirs
    # dirs have files per piece of info we're keeping, e.g. repoid, install
    # reason, status, etc.

    def __init__(self, db_path):
        self.conf = misc.GenericHolder()
        self.conf.db_path = db_path
        self.conf.writable = False

        self._packages = {} # pkgid = dir
        if not os.path.exists(self.conf.db_path):
            try:
                _makedirs_no_umask(self.conf.db_path)
                self.conf.writable = True
            except (IOError, OSError) as e:
                # some sort of useful thing here? A warning?
                pass
        else:
            if os.access(self.conf.db_path, os.W_OK):
                self.conf.writable = True
        self.yumdb_cache = {'attr' : {}}

    def _get_dir_name(self, pkgtup, pkgid):
        if pkgid in self._packages:
            return self._packages[pkgid]
        (n, a, e, v,r) = pkgtup
        n = _sanitize(n)
        str_pkgid = pkgid
        if pkgid is None:
            str_pkgid = '<nopkgid>'
        elif dnf.pycomp.is_py2str_py3bytes(pkgid):
            str_pkgid = pkgid.decode()
        thisdir = '%s/%s/%s-%s-%s-%s-%s' % (self.conf.db_path,
                                            n[0], str_pkgid, n, v, r, a)
        self._packages[pkgid] = thisdir
        return thisdir

    def get_package(self, po=None, pkgtup=None, pkgid=None):
        """Return an RPMDBAdditionalDataPackage Object for this package"""
        if po:
            thisdir = self._get_dir_name(po.pkgtup, po._pkgid)
        elif pkgtup and pkgid:
            thisdir = self._get_dir_name(pkgtup, pkgid)
        else:
            raise ValueError("Missing arguments.")

        return RPMDBAdditionalDataPackage(self.conf, thisdir,
                                          yumdb_cache=self.yumdb_cache)


class RPMDBAdditionalDataPackage(object):

    # We do auto hardlink on these attributes
    _auto_hardlink_attrs = set(['checksum_type', 'reason',
                                'installed_by', 'changed_by',
                                'from_repo', 'from_repo_revision',
                                'from_repo_timestamp', 'releasever',
                                'command_line'])

    def __init__(self, conf, pkgdir, yumdb_cache=None):
        self._conf = conf
        self._mydir = pkgdir

        self._read_cached_data = {}

        #  'from_repo' is the most often requested piece of data, and is often
        # the same for a huge number of packages. So we use hardlinks to share
        # data, and try to optimize for that.
        #  It's useful for other keys too (installed_by/changed_by/reason/etc.)
        # so we make it generic.
        self._yumdb_cache = yumdb_cache

    def _auto_cache(self, attr, value, fn, info=None):
        """ Create caches for the attr. We have a per. object read cache so at
            worst we only have to read a single attr once. Then we expand that
            with (dev, ino) cache, so hardlink data can be read once for
            multiple packages. """
        self._read_cached_data[attr] = value
        if self._yumdb_cache is None:
            return

        nlinks = 1
        if info is not None:
            nlinks = info.st_nlink
        if nlinks <= 1 and attr not in self._auto_hardlink_attrs:
            return

        if value in self._yumdb_cache['attr']:
            sinfo = self._yumdb_cache['attr'][value][1]
            if info is not None and sinfo is not None:
                if (info.st_dev, info.st_ino) == (sinfo.st_dev, sinfo.st_ino):
                    self._yumdb_cache['attr'][value][2].add(fn)
                    self._yumdb_cache[fn] = value
                    return
            if self._yumdb_cache['attr'][value][0] >= nlinks:
                # We already have a better cache file.
                return

        self._yumdb_cache['attr'][value] = (nlinks, info, set([fn]))
        self._yumdb_cache[fn]            = value

    def _unlink_yumdb_cache(self, fn):
        """ Remove old values from the link cache. """
        if fn in self._yumdb_cache:
            ovalue = self._yumdb_cache[fn]
            if ovalue in self._yumdb_cache['attr']:
                self._yumdb_cache['attr'][ovalue][2].discard(fn)
                if not self._yumdb_cache['attr'][ovalue][2]:
                    del self._yumdb_cache['attr'][ovalue]
            del self._yumdb_cache[fn]

    def _link_yumdb_cache(self, fn, value):
        """ If we have a matching yumdb cache, link() to it instead of having
            to open()+write(). """
        if self._yumdb_cache is None:
            return False

        self._unlink_yumdb_cache(fn)

        if value not in self._yumdb_cache['attr']:
            return False

        assert self._yumdb_cache['attr'][value][2]
        try:
            lfn = next(iter(self._yumdb_cache['attr'][value][2]))
            misc.unlink_f(fn + '.tmp')
            os.link(lfn, fn + '.tmp')
            os.rename(fn + '.tmp', fn)
        except:
            return False

        self._yumdb_cache['attr'][value][2].add(fn)
        self._yumdb_cache[fn] = value

        return True

    def _attr2fn(self, attr):
        """ Given an attribute, return the filename. """
        return os.path.normpath(self._mydir + '/' + attr)

    def _write(self, attr, value):
        # check for self._conf.writable before going on?
        if not os.path.exists(self._mydir):
            _makedirs_no_umask(self._mydir)

        attr = _sanitize(attr)
        if attr in self._read_cached_data:
            del self._read_cached_data[attr]
        fn = self._attr2fn(attr)

        if attr.endswith('.tmp'):
            raise AttributeError("Cannot set attribute %s on %s" % (attr, self))

        # Auto hardlink some of the attrs...
        if self._link_yumdb_cache(fn, value):
            return

        # Default write()+rename()... hardlink -c can still help.
        misc.unlink_f(fn + '.tmp')

        fo = _open_no_umask(fn + '.tmp', 'w')
        try:
            dnf.pycomp.write_to_file(fo, value)
        except (OSError, IOError) as e:
            logger.error("Cannot set attribute %s on %s due to:  %s" % (attr, self, e.strerror))

        fo.flush()
        fo.close()
        del fo
        os.rename(fn +  '.tmp', fn) # even works on ext4 now!:o

        self._auto_cache(attr, value, fn)

    def _read(self, attr):
        attr = _sanitize(attr)

        if attr in self._read_cached_data:
            return self._read_cached_data[attr]
        fn = self._attr2fn(attr)

        if attr.endswith('.tmp'):
            raise AttributeError("%s has no attribute %s" % (self, attr))

        info = misc.stat_f(fn, ignore_EACCES=True)
        if info is None:
            raise AttributeError("%s has no attribute %s" % (self, attr))

        if info.st_nlink > 1 and self._yumdb_cache is not None:
            key = (info.st_dev, info.st_ino)
            if key in self._yumdb_cache:
                self._auto_cache(attr, self._yumdb_cache[key], fn, info)
                return self._read_cached_data[attr]

        fo, e = _iopen(fn)
        if fo is None:
            msg = '{}: {}'.format(type(e).__name__, dnf.i18n.ucd(e))
            msg = "For {} cannot open attribute {} due to {}".format(self, attr, msg)
            logger.debug(msg)
            raise AttributeError(msg)

        try:
            value = fo.read()
        except Exception as e:
            msg = '{}: {}'.format(type(e).__name__, dnf.i18n.ucd(e))
            msg = "For {} cannot open attribute {} due to {}".format(self, attr, msg)
            logger.debug(msg)
            raise AttributeError(msg)
        finally:
            fo.close()
            del fo

        if info.st_nlink > 1 and self._yumdb_cache is not None:
            self._yumdb_cache[key] = value
        self._auto_cache(attr, value, fn, info)

        return value

    def _delete(self, attr):
        """remove the attribute file"""

        attr = _sanitize(attr)
        fn = self._attr2fn(attr)
        if attr in self._read_cached_data:
            del self._read_cached_data[attr]
        self._unlink_yumdb_cache(fn)
        if os.path.exists(fn):
            try:
                os.unlink(fn)
            except (IOError, OSError) as e:
                logger.error("Cannot delete attribute %s on %s at %s due to: %s" % (attr, self, fn, e.strerror))

    def __getattr__(self, attr):
        try:
            res = self._read(attr)
        except AttributeError:
            return None
        return res

    def __setattr__(self, attr, value):
        if not attr.startswith('_'):
            if value is not None:
                self._write(attr, value)
        else:
            object.__setattr__(self, attr, value)

    def __delattr__(self, attr):
        if not attr.startswith('_'):
            self._delete(attr)
        else:
            object.__delattr__(self, attr)

    def __contains__(self, attr):
        #  This is faster than __iter__ and it makes things fail in a much more
        # obvious way in weird FS corruption cases like: BZ 593436
        x = self.get(attr)
        return x is not None

    def __iter__(self, show_hidden=False):
        for item in self._read_cached_data:
            yield item
        for item in glob.glob(self._mydir + '/*'):
            item = item[(len(self._mydir) + 1):]
            if item in self._read_cached_data:
                continue
            if not show_hidden and item.endswith('.tmp'):
                continue
            yield item

    def clean(self):
        # purge out everything
        for item in self.__iter__(show_hidden=True):
            self._delete(item)
        try:
            os.rmdir(self._mydir)
        except OSError:
            pass

    def get(self, attr, default=None):
        """retrieve an add'l data obj"""

        try:
            res = self._read(attr)
        except AttributeError:
            return default
        return res



# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from tests import support
import dnf
import itertools

class UpgradeTo(support.ResultTestCase):
    def test_upgrade_to(self):
        base = support.MockBase("main", "updates")
        sack = base.sack
        base.upgrade("pepper-20-1.x86_64")
        new_set = support.installed_but(sack, "pepper").run()
        q = sack.query().available()._nevra("pepper-20-1.x86_64")
        new_set.extend(q)
        self.assertResult(base, new_set)

    def test_upgrade_to_reponame(self):
        """Test whether only packages in selected repo are used."""
        base = support.MockBase('updates', 'third_party')
        base.init_sack()

        base.upgrade('hole-1-2.x86_64', 'updates')

        self.assertResult(base, itertools.chain(
            base.sack.query().installed().filter(name__neq='hole'),
            dnf.subject.Subject('hole-1-2.x86_64').get_best_query(base.sack).filter(reponame='updates')))

    def test_upgrade_to_reponame_not_in_repo(self):
        """Test whether no packages are upgraded if bad repo is selected."""
        base = support.MockBase('main', 'updates')
        base.init_sack()

        base.upgrade('hole-1-2.x86_64', 'main')

        self.assertResult(base, base.sack.query().installed())



# Copyright (C) 2017 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import os
import shutil
import tempfile
import unittest

import hawkey
from hawkey import ModuleForm

import dnf.conf
from dnf.module.subject import ModuleSubject
from dnf.module.repo_module_dict import RepoModuleDict
from dnf.module.repo_module_version import RepoModuleVersion

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

TOP_DIR = os.path.abspath(os.path.dirname(__file__))
MODULES_DIR = os.path.join(TOP_DIR, "modules/etc/dnf/modules.d")
REPOS_DIR = os.path.join(TOP_DIR, "modules/modules")
DEFAULTS_DIR = os.path.join(TOP_DIR, "modules/etc/dnf/modules.defaults.d")

# with profile
MODULE_NSVAP = "module-name:stream:1::x86_64/profile"
MODULE_NSVP = "module-name:stream:1/profile"
MODULE_NSAP = "module-name:stream::x86_64/profile"
MODULE_NSP = "module-name:stream/profile"
MODULE_NP = "module-name/profile"
MODULE_NAP = "module-name::x86_64/profile"

# without profile
MODULE_NSVA = "module-name:stream:1::x86_64"
MODULE_NSV = "module-name:stream:1"
MODULE_NSA = "module-name:stream::x86_64"
MODULE_NS = "module-name:stream"
MODULE_N = "module-name"
MODULE_NA = "module-name::x86_64"


class ModuleSubjectTest(unittest.TestCase):
    def test_nsvap(self):
        subj = ModuleSubject(MODULE_NSVAP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSVAP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch="x86_64", profile="profile")
        self.assertEqual(actual, expected)

    def test_nsva(self):
        subj = ModuleSubject(MODULE_NSVA)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSVA))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch="x86_64", profile=None)
        self.assertEqual(actual, expected)

        # empty profile spec -> no profile
        subj = ModuleSubject(MODULE_NSVA + "/")
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSVA))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch="x86_64", profile=None)
        self.assertEqual(actual, expected)

    def test_nsvp(self):
        subj = ModuleSubject(MODULE_NSVAP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSVAP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch="x86_64", profile="profile")
        self.assertEqual(actual, expected)

    def test_nsv(self):
        subj = ModuleSubject(MODULE_NSV)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSV))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch=None, profile=None)
        self.assertEqual(actual, expected)

    def test_nsap(self):
        subj = ModuleSubject(MODULE_NSAP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSAP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=None,
                              arch="x86_64", profile="profile")
        self.assertEqual(actual, expected)

    def test_nsa(self):
        subj = ModuleSubject(MODULE_NSA)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSA))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=None,
                              arch="x86_64", profile=None)
        self.assertEqual(actual, expected)

    def test_nsp(self):
        subj = ModuleSubject(MODULE_NSP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NSP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=None,
                              arch=None, profile="profile")
        self.assertEqual(actual, expected)

    def test_ns(self):
        subj = ModuleSubject(MODULE_NS)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NS))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=None,
                              arch=None, profile=None)
        self.assertEqual(actual, expected)

    def test_nap(self):
        subj = ModuleSubject(MODULE_NAP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NAP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream=None, version=None,
                              arch="x86_64", profile="profile")
        self.assertEqual(actual, expected)

    def test_na(self):
        subj = ModuleSubject(MODULE_NA)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NA))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream=None, version=None,
                              arch="x86_64", profile=None)
        self.assertEqual(actual, expected)

    def test_np(self):
        subj = ModuleSubject(MODULE_NP)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_NP))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream=None, version=None,
                              arch=None, profile="profile")
        self.assertEqual(actual, expected)

    def test_n(self):
        subj = ModuleSubject(MODULE_N)
        result = list(subj.get_module_form_possibilities(forms=hawkey.MODULE_FORM_N))
        self.assertEqual(len(result), 1)
        actual = result[0]
        expected = ModuleForm(name="module-name", stream=None, version=None,
                              arch=None, profile=None)
        self.assertEqual(actual, expected)

    def test_all(self):
        subj = ModuleSubject(MODULE_NSVAP)
        result = list(subj.get_module_form_possibilities())
        self.assertEqual(len(result), 1)

        actual = result[0]
        expected = ModuleForm(name="module-name", stream="stream", version=1,
                              arch="x86_64", profile="profile")
        self.assertEqual(actual, expected)


class RepoModuleDictTest(unittest.TestCase):
    @staticmethod
    def _create_mmd(name, stream, version, rpms=None, profiles=None):
        rpms = rpms or []
        profiles = profiles or {}  # profile_name: {pkg_format: [pkg_names]}

        mmd = Modulemd.Module()
        mmd.set_name(str(name))
        mmd.set_stream(str(stream))
        mmd.set_version(int(version))
        licenses = Modulemd.SimpleSet()
        licenses.add("LGPLv2")
        mmd.set_module_licenses(licenses)
        mmd.set_summary(str("Fake module"))
        mmd.set_description(mmd.get_summary())
        for rpm in rpms:
            mmd.get_rpm_components.add(rpm.rsplit("-", 2)[0], "")
            mmd.get_rpm_artifacts.add(rpm[:-4])
        for profile_name in profiles:

            profile = Modulemd.Profile()
            for rpm in profiles[profile_name].get("rpms", []):
                profile.add_rpm(rpm)
            profile.set_name(profile_name)
            mmd.add_profile(profile)
        return mmd

    def test_find_module_version(self):
        rmd = RepoModuleDict(None)

        mmd = self._create_mmd(name="module-name", stream="stream", version=1,
                               profiles={"default": {}})
        rmv = RepoModuleVersion(mmd, None, None)
        rmd.add(rmv)

        mmd = self._create_mmd(name="module-name", stream="stream", version=2,
                               profiles={"default": {}})
        rmv = RepoModuleVersion(mmd, None, None)
        rmd.add(rmv)

        mmd = self._create_mmd(name="module-name", stream="enabled_stream", version=1,
                               profiles={"default": {}})
        rmv = RepoModuleVersion(mmd, None, None)
        rmd.add(rmv)

        mmd = self._create_mmd(name="module-name", stream="default_stream", version=1,
                               profiles={"default": {}})
        rmv = RepoModuleVersion(mmd, None, None)
        rmd.add(rmv)

        # set defaults
        defaults = Modulemd.Defaults()
        defaults.set_module_name("module-name")
        defaults.set_default_stream("stream")
        # no default profiles provided
        rmd["module-name"].defaults = defaults

        # no default, no active -> can't find stream automatically
        rmv = rmd.find_module_version(name="module-name")
        self.assertEqual(rmv.full_version, "module-name:stream:2")

        # set enabled stream
        conf = dnf.conf.ModuleConf()
        conf.enabled = 1
        conf.stream = "enabled_stream"
        rmd["module-name"].conf = conf

        # stream provided by user
        rmv = rmd.find_module_version(name="module-name", stream="stream")
        self.assertEqual(rmv.full_version, "module-name:stream:2")

        # stream and version provided by user
        rmv = rmd.find_module_version(name="module-name", stream="stream", version=1)
        self.assertEqual(rmv.full_version, "module-name:stream:1")

        # stream == active stream
        rmv = rmd.find_module_version(name="module-name")
        self.assertEqual(rmv.full_version, "module-name:enabled_stream:1")

        # stream == default stream
        conf.enabled = 0
        rmv = rmd.find_module_version(name="module-name")
        self.assertEqual(rmv.full_version, "module-name:stream:2")


class ModuleTest(unittest.TestCase):
    def assertInstalls(self, nevras):
        expected = sorted(set(nevras))
        actual = sorted(set([str(i) for i in self.base._goal.list_installs()]))
        self.assertEqual(expected, actual)

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="dnf_test_")
        self.conf = dnf.conf.Conf()
        self.conf.cachedir = os.path.join(self.tmpdir, "cache")
        self.conf.installroot = os.path.join(self.tmpdir, "root")
        self.conf.modulesdir = MODULES_DIR
        self.conf.moduledefaultsdir = DEFAULTS_DIR
        self.conf.substitutions["arch"] = "x86_64"
        self.conf.substitutions["basearch"] = dnf.rpm.basearch(self.conf.substitutions["arch"])
        self.conf.assumeyes = True
        self.base = dnf.Base(conf=self.conf)

        self._add_module_repo("_all")
        self.base.fill_sack(load_system_repo=False)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def _add_module_repo(self, repo_id, modules=True):
        url = "file://" + os.path.join(REPOS_DIR, repo_id, self.conf.substitutions["arch"])
        repo = self.base.repos.add_new_repo(repo_id, self.base.conf, baseurl=[url], modules=modules)
        return repo

    def test_module_dir_exist(self):
        dir = self.base.repo_module_dict.get_modules_dir()
        self.assertTrue(os.path.exists(dir))

    # dnf module enable

    def test_enable_name(self):
        # use default stream
        self.base.repo_module_dict.enable("httpd")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_enable_name_stream(self):
        self.base.repo_module_dict.enable("httpd:2.4")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        # also enable base-runtime; it's a dependency that's used in other tests
        self.base.repo_module_dict.enable("base-runtime:f26")

    def test_enable_pkgspec(self):
        self.base.repo_module_dict.enable("httpd:2.4:1/foo")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_enable_invalid(self):
        with self.assertRaises(dnf.exceptions.Error):
            self.base.repo_module_dict.enable("httpd:invalid")

    def test_enable_different_stream(self):
        repo_module = self.base.repo_module_dict["httpd"]

        self.base.repo_module_dict.enable("httpd:2.4")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        self.base.repo_module_dict.enable("httpd:2.2")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.2")

    def test_enable_different_stream_missing_profile(self):
        pass

    # dnf module disable

    def test_disable_name(self):
        repo_module = self.base.repo_module_dict["httpd"]

        self.base.repo_module_dict.enable("httpd")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        self.base.repo_module_dict.disable("httpd")
        self.assertFalse(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_disable_name_stream(self):
        repo_module = self.base.repo_module_dict["httpd"]

        self.base.repo_module_dict.enable("httpd:2.4")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        self.base.repo_module_dict.disable("httpd:2.4")
        self.assertFalse(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_disable_pkgspec(self):
        repo_module = self.base.repo_module_dict["httpd"]

        self.base.repo_module_dict.enable("httpd:2.4")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        self.base.repo_module_dict.disable("httpd:2.4:1/foo")
        self.assertFalse(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_disable_invalid(self):
        repo_module = self.base.repo_module_dict["httpd"]

        self.base.repo_module_dict.enable("httpd:2.4")
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

        with self.assertRaises(dnf.exceptions.Error):
            self.base.repo_module_dict.disable("httpd:invalid")

    # dnf module lock

    def test_lock_name(self):
        self.base.repo_module_dict.enable("httpd")
        self.base.repo_module_dict.lock("httpd")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_lock_name_stream(self):
        self.base.repo_module_dict.enable("httpd:2.4")
        self.base.repo_module_dict.lock("httpd:2.4")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_lock_pkgspec(self):
        self.base.repo_module_dict.enable("httpd:2.4:1/foo")
        self.base.repo_module_dict.lock("httpd:2.4:1/foo")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertTrue(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")

    def test_lock_invalid(self):
        with self.assertRaises(dnf.exceptions.Error):
            self.base.repo_module_dict.lock("httpd:invalid")

    # dnf module unlock

    def test_unlock_name(self):
        self.base.repo_module_dict.enable("httpd")
        self.base.repo_module_dict.lock("httpd")

        self.base.repo_module_dict.unlock("httpd")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertFalse(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")
        self.assertEqual(repo_module.conf.version, 2)

    def test_unlock_name_stream(self):
        self.base.repo_module_dict.enable("httpd:2.4")
        self.base.repo_module_dict.lock("httpd:2.4")

        self.base.repo_module_dict.unlock("httpd:2.4")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertFalse(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")
        self.assertEqual(repo_module.conf.version, 2)

    def test_unlock_pkgspec(self):
        self.base.repo_module_dict.enable("httpd:2.4:1/foo")
        self.base.repo_module_dict.lock("httpd:2.4:1/foo")

        self.base.repo_module_dict.unlock("httpd:2.4:1/foo")
        repo_module = self.base.repo_module_dict["httpd"]
        self.assertFalse(repo_module.conf.locked)
        self.assertEqual(repo_module.conf.name, "httpd")
        self.assertEqual(repo_module.conf.stream, "2.4")
        self.assertEqual(repo_module.conf.version, 1)

    def test_unlock_invalid(self):
        with self.assertRaises(dnf.exceptions.Error):
            self.base.repo_module_dict.unlock("httpd:invalid")

    # dnf module info

    def test_info_name(self):
        pass

    def test_info_name_stream(self):
        pass

    def test_info_pkgspec(self):
        pass

    # dnf module list

    def test_list(self):
        # show latest module versions
        rmd = self.base.repo_module_dict
        latest = rmd.list_module_version_latest()

        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertIn(rmv, latest)

        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=1)
        self.assertNotIn(rmv, latest)

    def test_list_all(self):
        rmd = self.base.repo_module_dict
        all = rmd.list_module_version_all()

        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertIn(rmv, all)

        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=1)
        self.assertIn(rmv, all)

    def test_list_enabled(self):
        rmd = self.base.repo_module_dict

        enabled = rmd.list_module_version_enabled()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertNotIn(rmv, enabled)

        rmd.enable("base-runtime:f26")

        enabled = rmd.list_module_version_enabled()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertIn(rmv, enabled)

    def test_list_disabled(self):
        rmd = self.base.repo_module_dict

        # enable
        rmd.enable("base-runtime:f26")

        # check not in disabled
        disabled = rmd.list_module_version_disabled()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertNotIn(rmv, disabled)

        # disable
        rmd.disable("base-runtime")

        # check in disabled
        disabled = rmd.list_module_version_disabled()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertIn(rmv, disabled)

    def test_list_installed(self):
        rmd = self.base.repo_module_dict

        # check not installed
        installed = rmd.list_module_version_installed()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertNotIn(rmv, installed)

        # install
        try:
            rmd.install(["base-runtime"])
        except SystemExit:
            # module profile wasn't, module was just enabled

            repo_module = rmd["base-runtime"]
            self.assertTrue(repo_module.conf.enabled)
            self.assertEqual(repo_module.conf.name, "base-runtime")
            self.assertEqual(repo_module.conf.stream, "f26")
            self.assertEqual(repo_module.conf.profiles, [])
            return

        # check module conf
        repo_module = rmd["base-runtime"]
        self.assertTrue(repo_module.conf.enabled)
        self.assertEqual(repo_module.conf.name, "base-runtime")
        self.assertEqual(repo_module.conf.stream, "f26")
        self.assertEqual(repo_module.conf.profiles, ["minimal"])

        # check installed
        installed = rmd.list_module_version_installed()
        rmv = rmd.find_module_version(name="base-runtime", stream="f26", version=2)
        self.assertIn(rmv, installed)

    # dnf module install / dnf install @

    def test_install_profile_latest(self):
        self.test_enable_name_stream()
        self.base.repo_module_dict.install(["httpd/default"])
        self.base.resolve()
        expected = [
            "basesystem-11-3.noarch",
            "filesystem-3.2-40.x86_64",
            "glibc-2.25.90-2.x86_64",
            "glibc-common-2.25.90-2.x86_64",
            "httpd-2.4.25-8.x86_64",
            "libnghttp2-1.21.1-1.x86_64",  # expected behaviour, non-modular rpm pulled in
        ]
        self.assertInstalls(expected)

    def test_install_profile(self):
        self.test_enable_name_stream()
        self.base.repo_module_dict.install(["httpd:2.4:1/default"])
        self.base.resolve()
        expected = [
            "basesystem-11-3.noarch",
            "filesystem-3.2-40.x86_64",
            "glibc-2.25.90-2.x86_64",
            "glibc-common-2.25.90-2.x86_64",
            "httpd-2.4.25-7.x86_64",
            "libnghttp2-1.21.1-1.x86_64",  # expected behaviour, non-modular rpm pulled in
        ]
        self.assertInstalls(expected)

    def test_install_two_profiles(self):
        self.test_enable_name_stream()

        self.base.repo_module_dict.install(["httpd:2.4:1/default", "httpd:2.4:1/doc"])
        self.base.resolve()
        expected = [
            "basesystem-11-3.noarch",
            "filesystem-3.2-40.x86_64",
            "glibc-2.25.90-2.x86_64",
            "glibc-common-2.25.90-2.x86_64",
            "httpd-2.4.25-7.x86_64",
            "httpd-doc-2.4.25-7.x86_64",
            "libnghttp2-1.21.1-1.x86_64",  # expected behaviour, non-modular rpm pulled in
        ]
        self.assertInstalls(expected)

    def test_install_two_profiles_different_versions(self):
        self.test_enable_name_stream()
        self.base.repo_module_dict.install(["httpd:2.4:2/default", "httpd:2.4:1/doc"])
        #self.base.resolve()
        self.base.resolve()
        expected = [
            "basesystem-11-3.noarch",
            "filesystem-3.2-40.x86_64",
            "glibc-2.25.90-2.x86_64",
            "glibc-common-2.25.90-2.x86_64",
            "httpd-2.4.25-8.x86_64",
            "httpd-doc-2.4.25-8.x86_64",
            "libnghttp2-1.21.1-1.x86_64",  # expected behaviour, non-modular rpm pulled in
        ]
        self.assertInstalls(expected)

    def test_install_profile_updated(self):
        return
        """
        # install profile1 from an old module version
        # then install profile2 from latest module version
        # -> dnf forces upgrade profile1 to the latest module version
        """

        self.test_install_profile()
        self.base.repo_module_dict.install(["httpd:2.4:2/doc"])
        self.base.resolve()
        expected = [
            "basesystem-11-3.noarch",
            "filesystem-3.2-40.x86_64",
            "glibc-2.25.90-2.x86_64",
            "glibc-common-2.25.90-2.x86_64",
            "httpd-2.4.25-8.x86_64",
            "httpd-doc-2.4.25-8.x86_64",
            "libnghttp2-1.21.1-1.x86_64",
        ]
        self.assertInstalls(expected)

    def test_install_deps_same_module_version(self):
        pass

    # dnf module upgrade / dnf upgrade @

    def test_upgrade(self):
        pass

    def test_upgrade_lower_rpm_nevra(self):
        pass

    def test_upgrade_lower_module_nsvap(self):
        pass

    def test_upgrade_missing_profile(self):
        pass

    # dnf module downgrade / dnf downgrade @

    def test_downgrade(self):
        pass

    # dnf module remove / dnf remove @

    def test_remove(self):
        pass

    def test_remove_shared_rpms(self):
        # don't remove RPMs that are part of another installed module / module profile
        # also don't remove RPMs that are required by user-installed RPMs
        pass

    def test_remove_invalid(self):
        pass

    def test_bare_rpms_filtering(self):
        """
        Test hybrid repos where RPMs of the same name (or Provides)
        can be both modular and bare (non-modular).
        """

        rmv = self.base.repo_module_dict.find_module_version(name="httpd")
        self.assertEqual(rmv.full_version, "httpd:2.4:2")

        # no match with modular RPM $name -> keep
        q = self.base.sack.query().filter(nevra="grub2-2.02-0.40.x86_64")
        self.assertEqual(len(q), 1)

        # $name matches with modular RPM $name -> exclude
        q = self.base.sack.query().filter(nevra="httpd-2.2.10-1.x86_64")
        self.assertEqual(len(q), 0)

        # Provides: $name matches with modular RPM $name -> exclude
        q = self.base.sack.query().filter(nevra="httpd-provides-name-3.0-1.x86_64")
        self.assertEqual(len(q), 0)

        # Provides: $name = ... matches with modular RPM $name -> exclude
        q = self.base.sack.query().filter(nevra="httpd-provides-name-version-release-3.0-1.x86_64")
        self.assertEqual(len(q), 0)



# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from tests import support
from tests.support import mock

import dnf.yum.rpmsack
import unittest


@mock.patch('os.path.exists', return_value=True)
class TestAdditionalPkgDB(unittest.TestCase):
    def _instantiate(self, base):
        path = base.conf.persistdir + '/yumdb'
        return dnf.yum.rpmsack.AdditionalPkgDB(path)

    def test_get_dir(self, mock_exists):
        base = support.MockBase()
        pkgdb = self._instantiate(base)
        pkg = base.sack.query().installed().filter(name="pepper")[0]
        expected = '%s/yumdb/p/bad9-pepper-20-0-x86_64' % base.conf.persistdir

        directory = pkgdb._get_dir_name(pkg.pkgtup, b'bad9')
        self.assertEqual(expected, directory)
        directory = pkgdb._get_dir_name(pkg.pkgtup, 'bad9')
        self.assertEqual(expected, directory)
        directory = pkgdb._get_dir_name(pkg.pkgtup, None)
        self.assertEqual('%s/yumdb/p/<nopkgid>-pepper-20-0-x86_64' %
                         base.conf.persistdir, directory)



# package.py
# Module defining the dnf.Package class.
#
# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

""" Contains the dnf.Package class. """

from __future__ import absolute_import
from __future__ import unicode_literals

import binascii
import dnf.rpm
import dnf.yum.misc
import hawkey
import logging
import os

logger = logging.getLogger("dnf")


class Package(hawkey.Package):
    """ Represents a package. #:api """

    def __init__(self, initobject, base):
        super(Package, self).__init__(initobject)
        self.base = base
        self._priv_chksum = None
        self._repo = None
        self._priv_size = None

    @property
    def _chksum(self):
        if self._priv_chksum:
            return self._priv_chksum
        if self._from_cmdline:
            chksum_type = dnf.yum.misc.get_default_chksum_type()
            chksum_val = dnf.yum.misc.checksum(chksum_type, self.location)
            return (hawkey.chksum_type(chksum_type),
                    binascii.unhexlify(chksum_val))
        return super(Package, self).chksum

    @_chksum.setter
    def _chksum(self, val):
        self._priv_chksum = val

    @property
    def debug_name(self):
        # :api
        """
        returns name of debuginfo package for given package
        e.g. kernel-PAE -> kernel-PAE-debuginfo
        """
        return "{}-debuginfo".format(self.name)

    @property
    def _from_cmdline(self):
        return self.reponame == hawkey.CMDLINE_REPO_NAME

    @property
    def _from_system(self):
        return self.reponame == hawkey.SYSTEM_REPO_NAME

    @property
    def _from_repo(self):
        yumdb_info = self.base._yumdb.get_package(self) if self._from_system else {}
        if 'from_repo' in yumdb_info:
            return '@'+yumdb_info.from_repo
        return self.reponame

    @property
    def _header(self):
        return dnf.rpm._header(self.localPkg())

    @property
    def _size(self):
        if self._priv_size:
            return self._priv_size
        return super(Package, self).size

    @_size.setter
    def _size(self, val):
        self._priv_size = val

    @property
    def source_debug_name(self):
        # :api
        """
        returns name of debuginfo package for source package of given package
        e.g. krb5-libs -> krb5-debuginfo
        """
        return "{}-debuginfo".format(self.source_name)

    @property
    def source_name(self):
        # :api
        """"
        returns name of source package
        e.g. krb5-libs -> krb5
        """
        if self.sourcerpm is not None:
            # trim suffix first
            srcname = dnf.util.rtrim(self.sourcerpm, ".src.rpm")
            # sourcerpm should be in form of name-version-release now, so we
            # will strip the two rightmost parts separated by dash.
            # Using rtrim with version and release of self is not sufficient
            # because the package can have different version to the source
            # package.
            srcname = srcname.rsplit('-', 2)[0]
        else:
            srcname = None
        return srcname

    @property
    def _pkgid(self):
        try:
            (_, chksum) = self.hdr_chksum
            return binascii.hexlify(chksum)
        except AttributeError:
            return None

    @property # yum compatibility attribute
    def idx(self):
        """ Always type it to int, rpm bindings expect it like that. """
        return int(self.rpmdbid)

    @property # yum compatibility attribute
    def repoid(self):
        return self.reponame

    @property # yum compatibility attribute
    def pkgtup(self):
        return (self.name, self.arch, str(self.e), self.v, self.r)

    @property # yum compatibility attribute
    def repo(self):
        if self._repo:
            return self._repo
        return self.base.repos[self.reponame]

    @repo.setter
    def repo(self, val):
        self._repo = val

    @property # yum compatibility attribute
    def relativepath(self):
        return self.location

    @property # yum compatibility attribute
    def a(self):
        return self.arch

    @property # yum compatibility attribute
    def e(self):
        return self.epoch

    @property # yum compatibility attribute
    def v(self):
        return self.version

    @property # yum compatibility attribute
    def r(self):
        return self.release

    @property # yum compatibility attribute
    def ui_from_repo(self):
        return self.reponame

    # yum compatibility method
    def evr_eq(self, pkg):
        return self.evr_cmp(pkg) == 0

    # yum compatibility method
    def evr_gt(self, pkg):
        return self.evr_cmp(pkg) > 0

    # yum compatibility method
    def evr_lt(self, pkg):
        return self.evr_cmp(pkg) < 0

    # yum compatibility method
    def getDiscNum(self):
        return self.medianr

    # yum compatibility method
    def localPkg(self):
        """ Package's location in the filesystem.

            For packages in remote repo returns where the package will be/has
            been downloaded.
        """
        if self._from_cmdline:
            return self.location
        loc = self.location
        if not self.repo._local:
            loc = os.path.basename(loc)
        elif self.baseurl and self.baseurl.startswith('file://'):
            return os.path.join(self.baseurl, loc)[7:]
        return os.path.join(self.repo.pkgdir, loc)

    def remote_location(self, schemes=('http', 'ftp', 'file', 'https')):
        # :api
        """
        The location from where the package can be downloaded from

        :param schemes: list of allowed protocols. Default is ('http', 'ftp', 'file', 'https')
        :return: location (string) or None
        """
        def schemes_filter(url_list):
            for url in url_list:
                if schemes:
                    s = dnf.pycomp.urlparse.urlparse(url)[0]
                    if s in schemes:
                        return os.path.join(url, self.location.lstrip('/'))
                else:
                    return os.path.join(url, self.location.lstrip('/'))
            return None

        if not self.location:
            return None

        if self.repo.metadata._mirrors:
            return schemes_filter(self.repo.metadata._mirrors)
        elif self.repo.baseurl:
            if isinstance(self.repo.baseurl, list):
                return schemes_filter(self.repo.baseurl)
            else:
                return schemes_filter([self.repo.baseurl])

    def _is_local_pkg(self):
        return self._from_cmdline or \
            (self.repo._local and (not self.baseurl or self.baseurl.startswith('file://')))

    # yum compatibility method
    def returnIdSum(self):
        """ Return the chksum type and chksum string how the legacy yum expects
            it.
        """
        (chksum_type, chksum) = self._chksum
        return (hawkey.chksum_name(chksum_type), binascii.hexlify(chksum).decode())

    # yum compatibility method
    def verifyLocalPkg(self):
        if self._from_system:
            raise ValueError("Can not verify an installed package.")
        if self._from_cmdline:
            return True # local package always verifies against itself
        (chksum_type, chksum) = self.returnIdSum()
        real_sum = dnf.yum.misc.checksum(chksum_type, self.localPkg(),
                                         datasize=self._size)
        if real_sum != chksum:
            logger.debug('%s: %s check failed: %s vs %s',
                         self, chksum_type, real_sum, chksum)
            return False
        return True



# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


class ModulePersistor(object):

    def __init__(self):
        self.repo_modules = []
        self._commit = False

    def set_data(self, repo_module, **kwargs):
        self.repo_modules.append(repo_module)
        for name, value in kwargs.items():
            setattr(repo_module.conf, name, value)

    def commit(self):
        self._commit = True

    def save(self):
        if not self._commit:
            return

        self._commit = False

        for repo_module in self.repo_modules:
            repo_module.write_conf_to_file()

        return True

    def reset(self):
        self.repo_modules = []
        self._commit = False



# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from collections import OrderedDict

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

from dnf.conf import ModuleConf
from dnf.module import module_messages, DIFFERENT_STREAM_INFO
from dnf.module.exceptions import NoStreamException, EnabledStreamException
from dnf.module.repo_module_stream import RepoModuleStream
from dnf.pycomp import ConfigParser
from dnf.util import logger


class RepoModule(OrderedDict):
    def __init__(self):
        super(RepoModule, self).__init__()

        self._conf = None
        self._defaults = None
        self.name = None
        self.parent = None

    @property
    def conf(self):
        if self._conf is None:
            self._conf = ModuleConf(section=self.name, parser=ConfigParser())
            self._conf.name = self.name
            self._conf.enabled = False
            self._conf.locked = False
            self._conf.version = -1

        return self._conf

    @conf.setter
    def conf(self, value):
        self._conf = value

    @property
    def defaults(self):
        if self._defaults is None:
            self._defaults = Modulemd.Defaults()
            self._defaults.set_module_name(self.name)
            # default stream and profiles remain unset
        return self._defaults

    @defaults.setter
    def defaults(self, value):
        self._defaults = value

    def add(self, repo_module_version):
        module_stream = self.setdefault(repo_module_version.stream, RepoModuleStream())
        module_stream.add(repo_module_version)
        module_stream.parent = self

        self.name = repo_module_version.name
        repo_module_version.repo_module = self

    def enable(self, stream, assumeyes=False):
        if stream not in self:
            raise NoStreamException("{}:{}".format(self.name, stream))

        if self.conf.enabled and self.conf.stream == stream:
            return

        if self.conf.stream is not None and \
                str(self.conf.stream) != str(stream) and \
                not assumeyes:
            logger.info(module_messages[DIFFERENT_STREAM_INFO].format(self.name))

            if not self.parent.base.conf.assumeno and \
                    self.parent.base.output.userconfirm():
                self.parent.base._module_persistor.set_data(self, version=-1, profiles=set())
                self.enable(stream, True)
            else:
                raise EnabledStreamException("{}:{}".format(self.name, stream))

        self.parent.base._module_persistor.set_data(self, stream=stream, enabled=True,
                                                    version=-1, profiles=set())

    def disable(self):
        self.parent.base._module_persistor.set_data(self, enabled=False, profiles=[])

    def lock(self, version):
        self.parent.base._module_persistor.set_data(self, locked=True, version=version)

    def unlock(self):
        self.parent.base._module_persistor.set_data(self, locked=False)

    def write_conf_to_file(self):
        output_file = os.path.join(self.parent.get_modules_dir(), "%s.module" % self.conf.name)

        with open(output_file, "w") as config_file:
            self.conf._write(config_file)



# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from argparse import Namespace
from tests import support
from tests.support import TestCase
from tests.support import mock

import dnf.cli.cli
import dnf.conf
import dnf.goal
import dnf.repo
import dnf.repodict
import os
import re

VERSIONS_OUTPUT = """\
  Installed: pepper-0:20-0.x86_64 at Thu Jan  1 00:00:00 1970
  Built    :  at Thu Jan  1 00:00:00 1970

  Installed: tour-0:5-0.noarch at Thu Jan  1 00:00:00 1970
  Built    :  at Thu Jan  1 00:00:00 1970
"""


class VersionStringTest(TestCase):
    def test_print_versions(self):
        base = support.MockBase()
        output = support.MockOutput()
        with mock.patch('sys.stdout') as stdout,\
                mock.patch('dnf.sack._rpmdb_sack', return_value=base.sack):
            dnf.cli.cli.print_versions(['pepper', 'tour'], base, output)
        written = ''.join([mc[1][0] for mc in stdout.method_calls
                           if mc[0] == 'write'])
        self.assertEqual(written, VERSIONS_OUTPUT)


@mock.patch('dnf.cli.cli.logger', new_callable=support.mock_logger)
class BaseCliTest(support.ResultTestCase):
    def setUp(self):
        self._base = dnf.cli.cli.BaseCli()
        self._base._sack = support.mock_sack('main', 'updates')
        self._base._goal = dnf.goal.Goal(self._base.sack)
        self._base.output.term = support.MockTerminal()
        self._base.downgrade_to = mock.Mock(wraps=self._base.downgrade_to)

    def test_downgradePkgs(self, logger):
        self._base.downgradePkgs(('tour',))

        self.assertEqual(self._base.downgrade_to.mock_calls, [mock.call('tour', strict=False)])
        self.assertEqual(logger.mock_calls, [])

    def test_downgradePkgs_notfound(self, logger):
        with self.assertRaises(dnf.exceptions.Error) as ctx:
            self._base.downgradePkgs(('non-existent',))
        self.assertEqual(str(ctx.exception), 'No packages marked for downgrade.')

        self.assertEqual(self._base.downgrade_to.mock_calls,
                         [mock.call('non-existent', strict=False)])
        self.assertEqual(logger.mock_calls,
                         [mock.call.info('No package %s available.',
                                         'non-existent')])

    @mock.patch('dnf.cli.cli._', dnf.pycomp.NullTranslations().ugettext)
    def test_downgradePkgs_notinstalled(self, logger):
        pkg = support.ObjectMatcher(dnf.package.Package, {'name': 'lotus'})

        with self.assertRaises(dnf.exceptions.Error) as ctx:
            self._base.downgradePkgs(('lotus',))
        self.assertEqual(str(ctx.exception), 'No packages marked for downgrade.')

        self.assertEqual(self._base.downgrade_to.mock_calls, [mock.call('lotus', strict=False)])



@mock.patch('dnf.cli.cli.Cli._read_conf_file')
class CliTest(TestCase):
    def setUp(self):
        self.base = support.MockBase("main")
        self.base.output = support.MockOutput()
        self.cli = dnf.cli.cli.Cli(self.base)

    def test_knows_upgrade(self, _):
        upgrade = self.cli.cli_commands['upgrade']
        update = self.cli.cli_commands['update']
        self.assertIs(upgrade, update)

    def test_simple(self, _):
        self.assertFalse(self.base.conf.assumeyes)
        self.cli.configure(['update', '-y'])
        self.assertTrue(self.base.conf.assumeyes)

    def test_glob_options_cmds(self, _):
        params = [
            ['install', '-y', 'pkg1', 'pkg2'],
            ['install', 'pkg1', '-y', 'pkg2'],
            ['install', 'pkg1', 'pkg2', '-y'],
            ['-y', 'install', 'pkg1', 'pkg2']
        ]
        for param in params:
            self.cli.configure(args=param)
            self.assertTrue(self.base.conf.assumeyes)
            self.assertEqual(self.cli.command.opts.command, ["install"])
            self.assertEqual(self.cli.command.opts.pkg_specs, ["pkg1", "pkg2"])

    def test_configure_repos(self, _):
        opts = Namespace()
        opts.repo = []
        opts.repos_ed = [('*', 'disable'), ('comb', 'enable')]
        opts.cacheonly = True
        opts.nogpgcheck = True
        opts.repofrompath = {}
        self.base._repos = dnf.repodict.RepoDict()
        self.base._repos.add(support.MockRepo('one', self.base.conf))
        self.base._repos.add(support.MockRepo('two', self.base.conf))
        self.base._repos.add(support.MockRepo('comb', self.base.conf))
        self.cli._configure_repos(opts)
        self.assertFalse(self.base.repos['one'].enabled)
        self.assertFalse(self.base.repos['two'].enabled)
        self.assertTrue(self.base.repos['comb'].enabled)
        self.assertFalse(self.base.repos["comb"].gpgcheck)
        self.assertFalse(self.base.repos["comb"].repo_gpgcheck)
        self.assertEqual(self.base.repos["comb"]._sync_strategy,
                         dnf.repo.SYNC_ONLY_CACHE)

    def test_configure_repos_expired(self, _):
        """Ensure that --cacheonly beats the expired status."""
        opts = Namespace()
        opts.repo = []
        opts.repos_ed = []
        opts.cacheonly = True
        opts.repofrompath = {}

        pers = self.base._repo_persistor
        pers.get_expired_repos = mock.Mock(return_value=('one',))
        self.base._repos = dnf.repodict.RepoDict()
        self.base._repos.add(support.MockRepo('one', self.base.conf))
        self.cli._configure_repos(opts)
        # _process_demands() should respect --cacheonly in spite of modified demands
        self.cli.demands.fresh_metadata = False
        self.cli.demands.cacheonly = True
        self.cli._process_demands()
        self.assertEqual(self.base.repos['one']._sync_strategy,
                         dnf.repo.SYNC_ONLY_CACHE)

@mock.patch('dnf.logging.Logging._setup', new=mock.MagicMock)
class ConfigureTest(TestCase):
    def setUp(self):
        self.base = support.MockBase("main")
        self.base._conf = dnf.conf.Conf()
        self.base.output = support.MockOutput()
        self.base._plugins = mock.Mock()
        self.cli = dnf.cli.cli.Cli(self.base)
        self.cli.command = mock.Mock()
        self.conffile = os.path.join(support.dnf_toplevel(), "etc/dnf/dnf.conf")

    @mock.patch('dnf.util.am_i_root', lambda: False)
    def test_configure_user(self):
        """ Test Cli.configure as user."""
        self.base._conf = dnf.conf.Conf()
        with mock.patch('dnf.rpm.detect_releasever', return_value=69):
            self.cli.configure(['update', '-c', self.conffile])
        reg = re.compile('^/var/tmp/dnf-[a-zA-Z0-9_-]+$')
        self.assertIsNotNone(reg.match(self.base.conf.cachedir))
        self.assertEqual(self.cli.cmdstring, "dnf update -c %s " % self.conffile)

    @mock.patch('dnf.util.am_i_root', lambda: True)
    def test_configure_root(self):
        """ Test Cli.configure as root."""
        self.base._conf = dnf.conf.Conf()
        with mock.patch('dnf.rpm.detect_releasever', return_value=69):
            self.cli.configure(['update', '--nogpgcheck', '-c', self.conffile])
        reg = re.compile('^/var/cache/dnf$')
        self.assertIsNotNone(reg.match(self.base.conf.system_cachedir))
        self.assertEqual(self.cli.cmdstring,
                         "dnf update --nogpgcheck -c %s " % self.conffile)

    def test_configure_verbose(self):
        with mock.patch('dnf.rpm.detect_releasever', return_value=69):
            self.cli.configure(['-v', 'update', '-c', self.conffile])
        self.assertEqual(self.cli.cmdstring, "dnf -v update -c %s " %
                         self.conffile)
        self.assertEqual(self.base.conf.debuglevel, 6)
        self.assertEqual(self.base.conf.errorlevel, 6)

    @mock.patch('dnf.cli.cli.Cli._parse_commands', new=mock.MagicMock)
    @mock.patch('os.path.exists', return_value=True)
    def test_conf_exists_in_installroot(self, ospathexists):
        with mock.patch('logging.Logger.warning') as warn, \
            mock.patch('dnf.rpm.detect_releasever', return_value=69):
            self.cli.configure(['--installroot', '/roots/dnf', 'update'])
        self.assertEqual(self.base.conf.config_file_path, '/roots/dnf/etc/dnf/dnf.conf')
        self.assertEqual(self.base.conf.installroot, '/roots/dnf')

    @mock.patch('dnf.cli.cli.Cli._parse_commands', new=mock.MagicMock)
    @mock.patch('os.path.exists', return_value=False)
    def test_conf_notexists_in_installroot(self, ospathexists):
        with mock.patch('dnf.rpm.detect_releasever', return_value=69):
            self.cli.configure(['--installroot', '/roots/dnf', 'update'])
        self.assertEqual(self.base.conf.config_file_path, '/etc/dnf/dnf.conf')
        self.assertEqual(self.base.conf.installroot, '/roots/dnf')

    @mock.patch('dnf.cli.cli.Cli._parse_commands', new=mock.MagicMock)
    def test_installroot_with_etc(self):
        """Test that conffile is detected in a new installroot."""
        self.cli.base.extcmds = []

        tlv = support.dnf_toplevel()
        self.cli.configure(['--installroot', tlv, 'update'])
        self.assertEqual(self.base.conf.config_file_path, '%s/etc/dnf/dnf.conf' % tlv)

    def test_installroot_configurable(self):
        """Test that conffile is detected in a new installroot."""

        conf = os.path.join(support.dnf_toplevel(), "tests/etc/installroot.conf")
        self.cli.configure(['-c', conf, '--nogpgcheck', '--releasever', '17', 'update'])
        self.assertEqual(self.base.conf.installroot, '/roots/dnf')



# transaction.py
# Managing the transaction to be passed to RPM.
#
# Copyright (C) 2013-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from dnf.i18n import _
from functools import reduce
import operator

DOWNGRADE = 1
ERASE     = 2
INSTALL   = 3
REINSTALL = 4
UPGRADE   = 5
FAIL      = 6


class TransactionItem(object):
    # :api

    __slots__ = ('op_type', 'installed', 'erased', 'obsoleted', 'reason')

    def __init__(self, op_type, installed=None, erased=None, obsoleted=None,
                 reason='unknown'):
        self.op_type = op_type
        self.installed = installed
        self.erased = erased
        self.obsoleted = list() if obsoleted is None else obsoleted
        self.reason = reason # reason for it to be in the transaction set

    @property
    def _active(self):
        return self.installed if self.installed is not None else self.erased

    @property
    def _active_history_state(self):
        return (self._installed_history_state if self.installed is not None
                else self._erased_history_state)

    @property
    def _erased_history_state(self):
        return self._HISTORY_ERASE_STATES[self.op_type]

    _HISTORY_INSTALLED_STATES = {
        DOWNGRADE : 'Downgrade',
        INSTALL   : 'Install',
        REINSTALL : 'Reinstall',
        UPGRADE   : 'Update'
        }

    _HISTORY_ERASE_STATES = {
        DOWNGRADE : 'Downgraded',
        ERASE     : 'Erase',
        REINSTALL : 'Reinstalled',
        UPGRADE   : 'Updated'
        }

    def _history_iterator(self):
        if self.installed is not None:
            yield(self.installed, self._installed_history_state)
        if self.erased is not None:
            yield(self.erased, self._erased_history_state)
        if self.obsoleted:
            yield(self.installed, self._obsoleting_history_state)
        for obs in self.obsoleted:
            yield(obs, self._obsoleted_history_state)

    @property
    def _installed_history_state(self):
        return self._HISTORY_INSTALLED_STATES[self.op_type]

    def installs(self):
        # :api
        return [] if self.installed is None else [self.installed]

    @property
    def _obsoleted_history_state(self):
        return 'Obsoleted'

    @property
    def _obsoleting_history_state(self):
        return 'Obsoleting'

    def _propagated_reason(self, yumdb, installonlypkgs_query):
        if self.reason == 'user':
            return self.reason
        if installonlypkgs_query.filter(name=self.installed.name):
            return 'user'
        if self.op_type in [DOWNGRADE, REINSTALL, UPGRADE]:
            previously = yumdb.get_package(self.erased).get('reason')
            if previously:
                return previously
        if self.obsoleted:
            reasons = set()
            for obs in self.obsoleted:
                reasons.add(yumdb.get_package(obs).get('reason'))
            if reasons:
                if 'user' in reasons:
                    return 'user'
                if 'group' in reasons:
                    return 'group'
                if 'dep' in reasons:
                    return 'dep'
                if 'weak' in reasons:
                    return 'weak'
        return self.reason

    def removes(self):
        # :api
        l =  [] if self.erased is None else [self.erased]
        return l + self.obsoleted


class Transaction(object):
    # :api

    def __init__(self):
        # :api
        self._tsis = []

    def __iter__(self):
        #: api
        return iter(self._tsis)

    def __len__(self):
        return len(self._tsis)

    def _items2set(self, extracting_fn):
        lists = map(extracting_fn, self._tsis)
        sets = map(set, lists)
        return reduce(operator.or_, sets, set())

    def add_downgrade(self, new, downgraded, obsoleted):
        # :api
        tsi = TransactionItem(DOWNGRADE, new, downgraded, obsoleted)
        self._tsis.append(tsi)

    def add_erase(self, erased):
        # :api
        tsi = TransactionItem(ERASE, erased=erased)
        self._tsis.append(tsi)

    def add_install(self, new, obsoleted, reason='unknown'):
        # :api
        tsi = TransactionItem(INSTALL, new, obsoleted=obsoleted,
                              reason=reason)
        self._tsis.append(tsi)

    def add_reinstall(self, new, reinstalled, obsoleted):
        # :api
        tsi = TransactionItem(REINSTALL, new, reinstalled, obsoleted)
        self._tsis.append(tsi)

    def add_upgrade(self, upgrade, upgraded, obsoleted):
        # :api
        tsi = TransactionItem(UPGRADE, upgrade, upgraded, obsoleted)
        self._tsis.append(tsi)

    def _get_items(self, op_type):
        return [tsi for tsi in self._tsis if tsi.op_type == op_type]

    @property
    def install_set(self):
        # :api
        fn = operator.methodcaller('installs')
        return self._items2set(fn)

    def _populate_rpm_ts(self, ts):
        """Populate the RPM transaction set."""

        for tsi in self._tsis:
            if tsi.op_type == DOWNGRADE:
                ts.addErase(tsi.erased.idx)
                hdr = tsi.installed._header
                ts.addInstall(hdr, tsi, 'u')
            elif tsi.op_type == ERASE:
                ts.addErase(tsi.erased.idx)
            elif tsi.op_type == INSTALL:
                hdr = tsi.installed._header
                if tsi.obsoleted:
                    ts.addInstall(hdr, tsi, 'u')
                else:
                    ts.addInstall(hdr, tsi, 'i')
            elif tsi.op_type == REINSTALL:
                # note: in rpm 4.12 there should not be set
                # rpm.RPMPROB_FILTER_REPLACEPKG to work
                ts.addReinstall(tsi.installed._header, tsi)
                if tsi.obsoleted:
                    for pkg in tsi.obsoleted:
                        ts.addErase(pkg.idx)
            elif tsi.op_type == UPGRADE:
                hdr = tsi.installed._header
                ts.addInstall(hdr, tsi, 'u')
        return ts

    @property
    def remove_set(self):
        # :api
        fn = operator.methodcaller('removes')
        return self._items2set(fn)

    def _rpm_limitations(self):
        """ Ensures all the members can be passed to rpm as they are to perform
            the transaction.
        """
        src_installs = [pkg for pkg in self.install_set if pkg.arch == 'src']
        if len(src_installs):
            return _("Will not install a source rpm package (%s).") % \
                src_installs[0]
        return None

    def _total_package_count(self):
        return len(self.install_set | self.remove_set)



# Copyright (C) 2017 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import os
import tempfile
import unittest

import dnf.conf


DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


class ModuleConfTest(unittest.TestCase):

    def setUp(self):
        self.dnf_conf = dnf.conf.Conf()
        self.dnf_conf.modulesdir = [os.path.join(DIR, "modules/etc/dnf/modules.d")]
        parser = dnf.conf.ConfigParser()
        self.conf = dnf.conf.ModuleConf(section="base-runtime", parser=parser)

        # name - equal to section name
        self.conf.stream = "f26"
        self.conf.version = 1
        # profiles - empty list by default
        self.conf.enabled = 1
        self.conf.locked = 0

    def test_options(self):
        self.assertEqual(self.conf.name, "base-runtime")
        self.assertEqual(self.conf.stream, "f26")
        self.assertEqual(self.conf.version, 1)
        self.assertEqual(self.conf.profiles, [])
        self.assertEqual(self.conf.enabled, True)
        self.assertEqual(self.conf.locked, False)

    def test_write(self):
        tmp_dir = tempfile.mkdtemp()
        tmp_file = os.path.join(tmp_dir, "%s.module" % self.conf._section)
        with open(tmp_file, "w") as config_file:
            self.conf._write(config_file)
        # TODO: compare
        os.unlink(tmp_file)
        os.rmdir(tmp_dir)



# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import hawkey

import dnf
from dnf.module.exceptions import NoProfileException, PossibleProfilesExceptions, \
    NoProfilesException
from dnf.subject import Subject


class RepoModuleVersion(object):
    def __init__(self, module_metadata, base, repo):
        self.module_metadata = module_metadata
        self.repo = repo
        self.base = base
        self.parent = None
        self.repo_module = None

    def __lt__(self, other):
        # for finding latest
        assert self.full_stream == other.full_stream
        if self.repo_module.conf.locked:
            return self.version < self.repo_module.conf.version
        return self.module_metadata.get_version() < other.module_metadata.get_version()

    def __repr__(self):
        return self.full_version

    def report_profile_error(self, profile, default_profiles_used=False):
        if default_profiles_used:
            raise NoProfileException("{}/{}".format(self.full_version, profile))
        elif self.profiles:
            raise PossibleProfilesExceptions("{}/{}".format(self.full_version, profile),
                                             self.profiles)
        else:
            raise NoProfilesException("{}/{}".format(self.full_version, profile))

    def install(self, profiles, default_profiles):
        result = self._install_profiles(profiles, False)
        if not profiles:
            result |= self._install_profiles(default_profiles, True)

        profiles.extend(self.repo_module.conf.profiles)
        profiles.extend(default_profiles)
        self.base._module_persistor.set_data(self.repo_module, stream=self.stream,
                                             version=self.version, profiles=sorted(set(profiles)))

        return result

    def _install_profiles(self, profiles, defaults_used):
        installed = self.base.sack.query().installed().run()
        installed_nevras = [str(pkg) for pkg in installed]

        result = False
        for profile in profiles:
            if profile not in self.profiles:
                self.report_profile_error(profile, defaults_used)

            for nevra_object in self.profile_nevra_objects(profile):
                nevr = self.nevra_object_to_nevr_str(nevra_object)
                nevra = "{}.{}".format(nevr, nevra_object.arch)

                if nevra not in installed_nevras:
                    self.base.install(nevr, reponame=self.repo.id, forms=hawkey.FORM_NEVR)
                    self.base._goal.group_members.add(nevra_object.name)
                    self.base._goal.module_members.add(nevra)
                    result = True

        return result

    def upgrade(self, profiles):
        installed = self.base.sack.query().installed().run()
        installed_nevras = [str(pkg) for pkg in installed]
        query_to_return = None

        for profile in profiles:
            if profile not in self.profiles:
                raise NoProfileException("{}/{}".format(self.full_version, profile))

            for nevra_object in self.profile_nevra_objects(profile):
                nevr = self.nevra_object_to_nevr_str(nevra_object)
                nevra = "{}.{}".format(nevr, nevra_object.arch)

                if nevra not in installed_nevras:
                    self.base.install(nevr, reponame=self.repo.id, forms=hawkey.FORM_NEVR)
                else:
                    # TODO: verify that filter(nevra) really works correctly
                    # (possibly breaks multilib)
                    query = self.base.sack.query().filter(nevra=nevra)
                    if query_to_return is None:
                        query_to_return = query
                    else:
                        query_to_return = query_to_return.union(query)

        self.base._module_persistor.set_data(self.repo_module, stream=self.stream,
                                             version=self.version)

        return query_to_return

    def remove(self, profiles):
        for profile in profiles:
            if profile not in self.profiles:
                raise NoProfileException("{}/{}".format(self.full_version, profile))

            for nevra_object in self.profile_nevra_objects(profile):
                nevr = self.nevra_object_to_nevr_str(nevra_object)
                remove_query = dnf.subject.Subject(nevr) \
                    .get_best_query(self.base.sack, forms=hawkey.FORM_NEVR)

                if not remove_query \
                        or self.base._yumdb.get_package(remove_query[0]).reason == 'user':
                    continue

                self.base._remove_if_unneeded(remove_query)

        conf = self.repo_module.conf
        version = conf.version
        profiles = [x for x in conf.profiles if x not in profiles]

        if len(conf.profiles) == 0:
            conf.version = -1

        self.base._module_persistor.set_data(self.repo_module, stream=self.stream, version=version,
                                             profiles=sorted(set(profiles)))

    def nevra(self):
        result = self.artifacts()
        # HACK: remove epoch to make filter(nevra=...) work
        result = [i.replace("-0:", "-") for i in result]
        return result

    def artifacts(self):
        return self.module_metadata.get_rpm_artifacts().get()

    def requires(self):
        requires = {}
        for dependencies in self.module_metadata.peek_dependencies():
            for name, streams in dependencies.peek_requires().items():
                requires[name] = streams.get()
        return requires

    def summary(self):
        return self.module_metadata.get_summary()

    def description(self):
        return self.module_metadata.get_description()

    @staticmethod
    def nevra_object_to_nevr_str(nevra_object):
        return "{}-{}".format(nevra_object.name, nevra_object.evr())

    def rpms(self, profile):
        return self.module_metadata.get_profiles()[profile].get_rpms().get()

    def profile_nevra_objects(self, profile):
        result = []
        rpms = set(self.rpms(profile))
        for nevra in self.nevra():
            subj = Subject(nevra)
            nevra_obj = list(subj.get_nevra_possibilities(hawkey.FORM_NEVRA))[0]
            if nevra_obj.name not in rpms:
                continue
            result.append(nevra_obj)
        return result

    @property
    def version(self):
        return self.module_metadata.get_version()

    @property
    def full_version(self):
        return "%s:%s:%s" % (
            self.module_metadata.get_name(), self.module_metadata.get_stream(),
            self.module_metadata.get_version())

    @property
    def stream(self):
        return self.module_metadata.get_stream()

    @property
    def full_stream(self):
        return "%s-%s" % (self.module_metadata.get_name(), self.module_metadata.get_stream())

    @property
    def name(self):
        return self.module_metadata.get_name()

    @property
    def profiles(self):
        return sorted(self.module_metadata.get_profiles())



# goal.py
# Customized hawkey.Goal
#
# Copyright (C) 2014-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from copy import deepcopy
from dnf.i18n import _
import logging
import hawkey

logger = logging.getLogger('dnf')

class Goal(hawkey.Goal):
    def __init__(self, sack):
        super(Goal, self).__init__(sack)
        self.group_members = set()
        self.module_members = set()
        self._installs = []

    def get_reason(self, pkg):
        code = super(Goal, self).get_reason(pkg)
        if code == hawkey.REASON_DEP:
            return 'dep'
        if code == hawkey.REASON_USER:
            if pkg.name in self.group_members:
                return 'group'
            return 'user'
        if code == hawkey.REASON_CLEAN:
            return 'clean'
        if code == hawkey.REASON_WEAKDEP:
            return 'weak'
        assert False, 'Unknown reason: %d' % code

    def group_reason(self, pkg, current_reason):
        if current_reason == 'unknown' and pkg.name in self.group_members:
            return 'group'
        return current_reason

    def install(self, *args, **kwargs):
        if args:
            self._installs.extend(args)
        if 'select' in kwargs:
            self._installs.extend(kwargs['select'].matches())
        return super(Goal, self).install(*args, **kwargs)

    def push_userinstalled(self, query, yumdb):
        msg = _('--> Finding unneeded leftover dependencies')
        logger.debug(msg)
        for pkg in query.installed():
            yumdb_info = yumdb.get_package(pkg)
            reason = getattr(yumdb_info, 'reason', 'user')
            if reason not in ('dep', 'weak'):
                self.userinstalled(pkg)

    def available_updates_diff(self, query):
        available_updates = set(query.upgrades().filter(arch__neq="src")
                                .latest().run())
        installable_updates = set(self.list_upgrades())
        installs = set(self.list_installs())
        return (available_updates - installable_updates) - installs



# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from collections import OrderedDict


class RepoModuleStream(OrderedDict):
    def __init__(self):
        super(RepoModuleStream, self).__init__()

        self.stream = None
        self.parent = None

    def add(self, repo_module_version):
        self[repo_module_version.version] = repo_module_version
        repo_module_version.parent = self

        self.stream = repo_module_version.stream

    def latest(self):
        return max(self.values())



# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import absolute_import
from __future__ import unicode_literals
from tests import support

import dnf.comps
import dnf.util
import operator
import warnings


class EmptyPersistorTest(support.ResultTestCase):
    """Test group operations with empty persistor."""

    def setUp(self):
        self.base = support.MockBase('main')
        self.base.read_mock_comps(False)
        self.base.init_sack()

    def test_group_install_exclude(self):
        comps = self.base.comps
        grp = comps.group_by_pattern('somerset')
        cnt = self.base.group_install(grp.id, ('optional',), exclude=('lotus',))
        self.assertEqual(cnt, 0)

    @support.mock.patch('locale.getlocale', return_value=('cs_CZ', 'UTF-8'))
    def test_group_install_locale(self, _unused):
        comps = self.base.comps
        grp = comps.group_by_pattern('Kritick\xe1 cesta (Z\xe1klad)')
        cnt = self.base.group_install(grp.id, ('mandatory',))
        self.assertEqual(cnt, 2)

    def test_group_install_exclude_glob(self):
        comps = self.base.comps
        grp = comps.group_by_pattern('somerset')
        cnt = self.base.group_install(grp.id, ('optional',), exclude=('lo*',))
        self.assertEqual(cnt, 0)

    def test_group_install_exclude_notexist(self):
        comps = self.base.comps
        grp = comps.group_by_pattern('somerset')
        cnt = self.base.group_install(grp.id, ('optional',), exclude=('x*',))
        self.assertEqual(cnt, 1)

    def test_finalize_comps_trans(self):
        trans = dnf.comps.TransactionBunch()
        trans.install = ('trampoline',)
        self.assertGreater(self.base._add_comps_trans(trans), 0)
        self.base._finalize_comps_trans()
        self.assertIn('trampoline', self.base._goal.group_members)
        (installed, removed) = self.installed_removed(self.base)
        self.assertCountEqual(map(str, installed), ('trampoline-2.1-1.noarch',))
        self.assertEmpty(removed)


class PresetPersistorTest(support.ResultTestCase):
    """Test group operations with some data in the persistor."""

    def setUp(self):
        self.base = support.MockBase("main")
        self.base.read_mock_comps()
        self.base.init_sack()

    def test_env_group_remove(self):
        cnt = self.base.env_group_remove(["sugar-desktop-environment"])
        self.assertEqual(3, cnt)
        with support.mock.patch('logging.Logger.error') as log:
            self.assertRaises(dnf.exceptions.Error,
                              self.base.env_group_remove,
                              ['nonexistent'])

    def test_environment_remove(self):
        prst = self.base._group_persistor
        env_ids = prst.environments_by_pattern("sugar-desktop-environment")
        self.assertEqual(env_ids, set(['sugar-desktop-environment']))
        env_id = dnf.util.first(env_ids)
        self.assertGreater(self.base.environment_remove(env_id), 0)
        p_env = prst.environment(env_id)
        self.assertFalse(p_env.installed)
        peppers = prst.group('Peppers')
        somerset = prst.group('somerset')
        self.assertFalse(peppers.installed)
        self.assertFalse(somerset.installed)

    def test_env_upgrade(self):
        prst = self.base._group_persistor
        cnt = self.base.environment_upgrade("sugar-desktop-environment")
        self.assertEqual(5, cnt)
        peppers = prst.group('Peppers')
        somerset = prst.group('somerset')
        self.assertTrue(peppers.installed)
        self.assertTrue(somerset.installed)

    def test_group_install(self):
        prst = self.base._group_persistor
        grp = self.base.comps.group_by_pattern('Base')
        p_grp = prst.group('base')
        self.assertFalse(p_grp.installed)

        self.assertEqual(self.base.group_install(grp.id, ('mandatory',)), 2)
        inst, removed = self.installed_removed(self.base)
        self.assertEmpty(inst)
        self.assertEmpty(removed)
        self.assertTrue(p_grp.installed)

    """
    this should be reconsidered once relengs document comps
    def test_group_install_broken(self):
        prst = self.base._group_persistor
        grp = self.base.comps.group_by_pattern('Broken Group')
        p_grp = prst.group('broken-group')
        self.assertFalse(p_grp.installed)

        self.assertRaises(dnf.exceptions.MarkingError,
                          self.base.group_install, grp.id,
                          ('mandatory', 'default'))
        p_grp = prst.group('broken-group')
        self.assertFalse(p_grp.installed)

        self.assertEqual(self.base.group_install(grp.id,
                                                 ('mandatory', 'default'),
                                                 strict=False), 1)
        inst, removed = self.installed_removed(self.base)
        self.assertLength(inst, 1)
        self.assertEmpty(removed)
        p_grp = prst.group('broken-group')
        self.assertTrue(p_grp.installed)
    """

    def test_group_remove(self):
        prst = self.base._group_persistor
        grp_ids = prst.groups_by_pattern('somerset')
        self.assertEqual(grp_ids, set(['somerset']))
        grp_id = dnf.util.first(grp_ids)
        p_grp = prst.group('somerset')

        self.assertGreater(self.base.group_remove(grp_id), 0)
        inst, removed = self.installed_removed(self.base)
        self.assertEmpty(inst)
        self.assertCountEqual([pkg.name for pkg in removed], ('pepper',))
        self.assertFalse(p_grp.installed)


class EnvironmentInstallTest(support.ResultTestCase):
    def setUp(self):
        """Set up a test where sugar is considered not installed."""
        self.base = support.MockBase("main")
        self.base.init_sack()
        self.base.read_mock_comps()
        self.prst = self.base._group_persistor
        p_env = self.prst.environment('sugar-desktop-environment')
        p_env.pkg_types = 0
        p_env.grp_types = 0
        del p_env.full_list[:]
        p_grp = self.prst.group('somerset')
        p_grp.pkg_types = 0
        del p_grp.full_list[:]

    def test_environment_install(self):
        comps = self.base.comps
        env = comps.environment_by_pattern("sugar-desktop-environment")
        self.base.environment_install(env.id, ('mandatory',))
        installed, _ = self.installed_removed(self.base)
        self.assertCountEqual(map(operator.attrgetter('name'), installed),
                              ('trampoline', 'lotus'))

        p_env = self.prst.environment('sugar-desktop-environment')
        self.assertCountEqual(p_env.full_list, ('somerset', 'Peppers'))
        self.assertTrue(p_env.installed)

        peppers = self.prst.group('Peppers')
        somerset = self.prst.group('somerset')
        self.assertTrue(all((peppers.installed, somerset.installed)))



# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import gzip

import os
import gi

from dnf.module.exceptions import LoadCacheException, MissingYamlException
from dnf.pycomp import PY3

gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd


class ModuleMetadataLoader(object):
    def __init__(self, repo=None):
        self.repo = repo

    @property
    def _metadata_fn(self):
        if self.repo.metadata:
            return self.repo.metadata._repo_dct.get("modules")

    def load(self):
        if self.repo is None:
            raise LoadCacheException(self.repo)

        yaml_file_path = None
        if not self._metadata_fn:
            repodata_dir = self.repo._cachedir + "/repodata/"
            files = os.listdir(repodata_dir)
            for file in files:
                if "modules.yaml" in file:
                    yaml_file_path = repodata_dir + file
                    break

        if not self._metadata_fn and not yaml_file_path:
            raise MissingYamlException(self.repo._cachedir)

        openfunc = open
        if (self._metadata_fn and self._metadata_fn.endswith('.gz')) \
                or (yaml_file_path and yaml_file_path.endswith('.gz')):
            openfunc = gzip.open
        with openfunc(self._metadata_fn or yaml_file_path, "r") as modules_yaml_fd:
            modules_yaml = modules_yaml_fd.read()

        if PY3 and isinstance(modules_yaml, bytes):
            modules_yaml = modules_yaml.decode("utf-8")

        return Modulemd.objects_from_string(modules_yaml)



# supplies the 'module' command.
#
# Copyright (C) 2014-2017  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import print_function

from dnf.cli import commands, CliError
from dnf.i18n import _
from dnf.module.exceptions import NoModuleException
from dnf.util import logger


class ModuleCommand(commands.Command):
    class SubCommand(commands.Command):

        def __init__(self, cli):
            super(ModuleCommand.SubCommand, self).__init__(cli)

    class ListSubCommand(SubCommand):

        aliases = ('list',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True

        def run_on_module(self):
            mods = self.base.repo_module_dict

            if self.opts.all:
                print(mods.get_brief_description_all(self.opts.module_nsvp))
            elif self.opts.enabled:
                print(mods.get_brief_description_enabled(self.opts.module_nsvp))
            elif self.opts.disabled:
                print(mods.get_brief_description_disabled(self.opts.module_nsvp))
            elif self.opts.installed:
                print(mods.get_brief_description_installed(self.opts.module_nsvp))
            else:
                print(mods.get_brief_description_latest(self.opts.module_nsvp))

            return 0

    class InfoSubCommand(SubCommand):

        aliases = ('info',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True

        def run_on_module(self):
            for spec in self.opts.module_nsvp:
                print()
                if self.opts.verbose:
                    print(self.base.repo_module_dict.get_full_info(spec))
                elif self.opts.profile:
                    print(self.base.repo_module_dict.get_info_profiles(spec))
                else:
                    print(self.base.repo_module_dict.get_info(spec))

    class EnableSubCommand(SubCommand):

        aliases = ('enable',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.root_user = True

        def run_on_module(self):
            for module_ns in self.opts.module_nsvp:
                self.base.repo_module_dict.enable(module_ns, True)
                logger.info("'{}' is enabled".format(module_ns))

    class DisableSubCommand(SubCommand):

        aliases = ('disable',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.root_user = True

        def run_on_module(self):
            for module_n in self.opts.module_nsvp:
                self.base.repo_module_dict.disable(module_n, True)
                logger.info("'{}' is disabled".format(module_n))

    class InstallSubCommand(SubCommand):

        aliases = ('install',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.resolving = True
            demands.root_user = True

        def run_on_module(self):
            module_specs = self.base.repo_module_dict.install(self.opts.module_nsvp)
            if module_specs:
                raise NoModuleException(", ".join(module_specs))

    class UpdateSubCommand(SubCommand):

        aliases = ('update',)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.resolving = True
            demands.root_user = True

        def run_on_module(self):
            module_specs, _, _ = self.base.repo_module_dict.upgrade(self.opts.module_nsvp, True)
            if module_specs:
                raise NoModuleException(", ".join(module_specs))

    class RemoveSubCommand(SubCommand):

        aliases = ('remove', 'erase',)

        def configure(self):
            demands = self.cli.demands
            demands.allow_erasing = True
            demands.available_repos = True
            demands.resolving = True
            demands.root_user = True
            demands.sack_activation = True

        def run_on_module(self):
            self.base.repo_module_dict.remove(self.opts.module_nsvp)

    class LockSubCommand(SubCommand):

        aliases = ("lock",)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.root_user = True

        def run_on_module(self):
            for module_n in self.opts.module_nsvp:
                stream, version = self.base.repo_module_dict.lock(module_n, True)
                logger.info("'{}' is locked (stream: {}, version: {})"
                            .format(module_n, stream, version))

    class UnlockSubCommand(SubCommand):

        aliases = ("unlock",)

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True
            demands.root_user = True

        def run_on_module(self):
            for module_n in self.opts.module_nsvp:
                self.base.repo_module_dict.unlock(module_n, True)
                logger.info("'{}' is unlocked".format(module_n))

    class ProfileInfoSubCommand(SubCommand):

        aliases = ("profile", "profile-info")

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True

        def run_on_module(self):
            for spec in self.opts.module_nsvp:
                print()
                print(self.base.repo_module_dict.get_info_profiles(spec))

    class StreamsSubCommand(SubCommand):

        aliases = ("streams", "enabled", "enabled-streams")

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True

        def run_on_module(self):
            logger.info(self.base.repo_module_dict
                        .get_brief_description_enabled(self.opts.module_nsvp))

    class ProvidesSubCommand(SubCommand):

        aliases = ("provides", )

        def configure(self):
            demands = self.cli.demands
            demands.available_repos = True
            demands.sack_activation = True

        def run_on_module(self):
            self.base.repo_module_dict.print_what_provides(self.opts.module_nsvp)

    SUBCMDS = {ListSubCommand, InfoSubCommand, EnableSubCommand,
               DisableSubCommand, InstallSubCommand, UpdateSubCommand,
               RemoveSubCommand, LockSubCommand, UnlockSubCommand,
               ProfileInfoSubCommand, StreamsSubCommand, ProvidesSubCommand}

    SUBCMDS_NOT_REQUIRED_ARG = {ListSubCommand, StreamsSubCommand}

    aliases = ("module",)
    summary = _("Interact with Modules.")

    def __init__(self, cli):
        super(ModuleCommand, self).__init__(cli)
        subcmd_objs = (subcmd(cli) for subcmd in self.SUBCMDS)
        self.subcmd = None
        self._subcmd_name2obj = {
            alias: subcmd for subcmd in subcmd_objs for alias in subcmd.aliases}

    def set_argparser(self, parser):
        subcommand_help = [subcmd.aliases[0] for subcmd in self.SUBCMDS]
        parser.add_argument('subcmd', nargs=1, choices=subcommand_help)
        parser.add_argument('module_nsvp', nargs='*')

        narrows = parser.add_mutually_exclusive_group()
        narrows.add_argument('--all', dest='all',
                             action='store_true',
                             help=_("show all modules"))
        narrows.add_argument('--enabled', dest='enabled',
                             action='store_true',
                             help=_("show only enabled modules"))
        narrows.add_argument('--disabled', dest='disabled',
                             action='store_true',
                             help=_("show only disabled modules"))
        narrows.add_argument('--installed', dest='installed',
                             action='store_true',
                             help=_("show only installed modules"))
        narrows.add_argument('--profile', dest='profile',
                             action='store_true',
                             help=_("show profile content"))
        narrows.add_argument('--autoenable', dest='autoenable',
                             action='store_true',
                             help=_("auto enable stream"))

    def configure(self):
        try:
            self.subcmd = self._subcmd_name2obj[self.opts.subcmd[0]]
        except (CliError, KeyError):
            self.cli.optparser.print_usage()
            raise CliError
        self.subcmd.opts = self.opts
        self.subcmd.configure()

    def run(self):
        self.check_required_argument()
        self.subcmd.run_on_module()

    def check_required_argument(self):
        not_required_argument = [alias
                                 for subcmd in self.SUBCMDS_NOT_REQUIRED_ARG
                                 for alias in subcmd.aliases]
        if self.opts.subcmd[0] not in not_required_argument:
            if not self.opts.module_nsvp:
                raise CliError(
                    "dnf {} {}: too few arguments".format(self.opts.command[0],
                                                          self.opts.subcmd[0]))
